# MIT License

# Copyright (c) 2023 Spill-Tea

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Primary Logging Manager, EpiLog."""

from __future__ import annotations

import logging
from typing import Set, Union


defaultFormat = logging.Formatter(
    " | ".join(
        [
            "%(asctime)s",
            "%(name)s",
            "%(levelname)s",
            "%(module)s.%(funcName)s:%(lineno)d",
            "%(message)s",
        ]
    )
)


def _check_level(level: int) -> bool:
    _levels: Set[int] = {
        logging.NOTSET,
        logging.DEBUG,
        logging.INFO,
        logging.WARN,
        logging.ERROR,
        logging.CRITICAL,
    }

    return level in _levels


class EpiLog:
    """Log Manager designed to centralize Local Module or Task level logging control.

    Args:
        level (int): Logging Level
        stream (logging.Handler): Where logs are written to
        formatter (logging.Formatter): Dictates how to logs are formatted

    Notes:
        * Natively Supports only a single Stream per instantiated logger.
        * Designed for local control of logging (i.e. Logging events from globally
            imported libraries are not captured)

    Examples:
        ``` python

            import logging
            from EpiLog import EpiLog

            formatter = logging.Formatter("%(asctime)s | %(levelname)s | %(message)s")
            manager = EpiLog(logging.DEBUG, formatter=formatter)
            log = manager.get_log(__name__)  # A logger created using manager settings
            log.debug("Logger Created")
        ```

    """

    __slots__ = ("_formatter", "_level", "_stream", "loggers")

    _level: int
    _stream: logging.Handler
    _formatter: logging.Formatter
    loggers: dict[str, logging.Logger]

    def __init__(
        self,
        level: int = logging.INFO,
        stream: Union[logging.Handler, None] = None,
        formatter: logging.Formatter = defaultFormat,
    ):
        self.loggers: dict[str, logging.Logger] = {}

        # Use property setters to manage attributes
        self.stream = stream or logging.StreamHandler()
        self.level = level
        self.formatter = formatter

    def __getitem__(self, item: str) -> logging.Logger:
        """Retrieve Logger by name."""
        return self.loggers[item]

    @property
    def level(self) -> int:
        """Logging Level."""
        return self._level

    @level.setter
    def level(self, value: int):
        """Set Logging Level."""
        if _check_level(value) is False:
            raise ValueError(f"Invalid Logging Level: {value}")

        self._level = value
        self.stream.setLevel(self.level)
        for log in self.loggers.values():
            log.setLevel(self.level)
            # Update the streams to reflect current level
            for handle in log.handlers:
                handle.setLevel(self.level)

    @property
    def formatter(self) -> logging.Formatter:
        """Logging Format."""
        return self._formatter

    @formatter.setter
    def formatter(self, value: Union[logging.Formatter, None]):
        """Set Logging Format."""
        if value is None:
            value = defaultFormat

        if not issubclass(value.__class__, logging.Formatter):
            raise TypeError(f"Incorrect Formatter Type: {value.__class__}")

        self._formatter: logging.Formatter = value
        self.stream.setFormatter(self.formatter)
        for log in self.loggers.values():
            for handle in log.handlers:
                handle.setFormatter(self.formatter)

    @property
    def stream(self) -> logging.Handler:
        """Stream Handler for Logging."""
        return self._stream

    @stream.setter
    def stream(self, value: Union[logging.Handler, None]):
        """Replace Logging Handler Streams."""
        if value is None:
            value = logging.StreamHandler()

        if not issubclass(value.__class__, (logging.Filterer, logging.Handler)):
            raise TypeError(f"Unsupported Stream Handler: {value.__class__}")

        if hasattr(self, "_stream"):
            previous = self._stream
            value.setFormatter(self.formatter)
            value.setLevel(self.level)
            self._stream: logging.Handler = value
            for log in self.loggers.values():
                log.removeHandler(previous)
                log.addHandler(self.stream)

        else:
            self._stream = value

    def remove(self, name: str | logging.Logger) -> None:
        """Remove a logger from local and global registry."""
        if isinstance(name, logging.Logger):
            name = name.name

        logging.Logger.manager.loggerDict.pop(name)
        log: logging.Logger = self.loggers.pop(name)

        for handler in log.handlers:
            handler.flush()
            if id(handler) == id(self.stream):
                continue
            handler.close()
        log.handlers.clear()

    def get_logger(self, name: str) -> logging.Logger:
        """Initialize a new logger."""
        log = logging.getLogger(name)
        log.setLevel(self.level)
        log.addHandler(self.stream)
        self.loggers[name] = log

        return log



"""Unit tests to examine the Expected Behavior of the EpiLog Logging Manager."""

import logging
import os
from io import StringIO
from typing import Callable, Union

import pytest

from EpiLog import EpiLog
from EpiLog.manager import defaultFormat


@pytest.mark.parametrize("names", [("a", "b", "c"), ("b", "c", "d", "e")])
def test_get_logger(names, build_manager: Callable[..., EpiLog]) -> None:
    """Tests that we correctly construct a named logger."""
    manager: EpiLog = build_manager()
    assert len(manager.loggers) == 0

    for n in names:
        manager.get_logger(n)
    assert len(manager.loggers) == len(names)
    assert all(i in manager.loggers for i in names)


def test_logging(build_manager: Callable[..., EpiLog]) -> None:
    """Makes certain no errors are raised while logging."""
    manager: EpiLog = build_manager()
    log = manager.get_logger("test")
    log.info("Bob boop beep")


def test_stream(build_manager: Callable[..., EpiLog]) -> None:
    """Tests that the message output by logging is actually written to stream."""
    stream = StringIO()
    handler = logging.StreamHandler(stream)
    manager: EpiLog = build_manager(stream=handler)

    log = manager.get_logger("test")
    message = "You are blind to reality and for that I am most proud"
    log.info(message)

    stream.seek(0)
    output = stream.read()
    stream.close()

    assert message in output, "Message not Found in output stream after logging"


def test_level_change(build_manager: Callable[..., EpiLog]) -> None:
    """Tests that modifying the Logging Level accurately filters log output."""
    stream = StringIO()
    handler = logging.StreamHandler(stream)

    manager: EpiLog = build_manager(level=logging.INFO, stream=handler)
    log: logging.Logger = manager.get_logger("test")

    message = (
        "I would say that he's blessedly unburdened with the complications of a"
        " university education."
    )
    log.debug(message)
    stream.seek(0)
    output = stream.read()
    assert message not in output, "Message should not have been Received"

    # Then repeat after Changing level
    manager.level = logging.DEBUG
    log.debug(message)

    stream.seek(0)
    output = stream.read()
    stream.close()

    assert message in output, "Message not Found in output stream after logging"


def test_format_change(build_manager: Callable[..., EpiLog]) -> None:
    """Tests that modifying the Logging Format accurately Modifies log output Format."""
    stream = StringIO()
    handler = logging.StreamHandler(stream)
    manager: EpiLog = build_manager(level=logging.INFO, stream=handler)
    log: logging.Logger = manager.get_logger("test")

    # Modify Formatter
    manager.formatter = logging.Formatter("%(levelname)s | %(message)s")
    message = "I have no idea why all of this is happening or how to control it."
    log.info(message)

    stream.seek(0)
    output = stream.read()
    stream.close()

    assert f"INFO | {message}\n" == output, "Expected Message format to match."


def test_stream_change(build_manager: Callable[..., EpiLog]) -> None:
    """Tests that modifying the Logging Stream writes to the correct Stream."""
    stream_a = StringIO()
    stream_b = StringIO()
    handler_a = logging.StreamHandler(stream_a)
    handler_b = logging.StreamHandler(stream_b)

    manager: EpiLog = build_manager(
        level=logging.INFO,
        stream=handler_a,
        formatter=logging.Formatter("%(levelname)s | %(message)s"),
    )
    log = manager.get_logger("test")

    # Modify Stream
    manager.stream = handler_b
    assert manager.stream == handler_b, "Unexpected Stream."

    message = (
        "You humans have so many emotions! You only need two: anger and confusion!"
    )
    expected = f"INFO | {message}\n"
    log.info(message)

    # Test first stream
    stream_a.seek(0)
    output_a = stream_a.read()
    stream_a.close()
    assert (
        output_a != expected
    ), "Did not Expect message to be written to previous stream."

    # Test Second, Expected stream
    stream_b.seek(0)
    output_b = stream_b.read()
    stream_b.close()
    assert output_b == expected, "Expected message to be written to current stream."


def test_stream_change_to_none(build_manager: Callable[..., EpiLog]) -> None:
    """Confirm that attempts to set the stream to none result in default stream."""
    manager: EpiLog = build_manager()
    assert manager.stream is not None, "Expected Default Stream to instantiate."
    previous_id = id(manager.stream)

    manager.stream = None  # type: ignore[assignment]
    assert manager.stream is not None, "Expected Default Stream to kick in."
    assert isinstance(
        manager.stream, (logging.Handler, logging.Filterer)
    ), "Unexpected stream type."

    assert id(manager.stream) != previous_id, "Expected new stream instance."


@pytest.mark.parametrize(
    "f",
    [
        None,
        logging.Formatter("%(name)s | %(levelname)s | %(message)s"),
    ],
)
def test_format(f, build_manager: Callable[..., EpiLog]) -> None:
    """Tests expected behaviors when instantiating with Formatters."""
    manager: EpiLog = build_manager(formatter=f)
    if f is not None:
        assert manager.formatter == f, "Expected Logging Format to reflect input."
    else:
        assert id(manager.formatter) == id(defaultFormat), "Expected default formatter"


def test_format_error(build_manager: Callable[..., EpiLog]) -> None:
    """Test TypeError is raised when instantiating with invalid Formatters."""
    with pytest.raises(TypeError):
        build_manager(formatter="Failure")


@pytest.mark.parametrize(
    "level",
    [
        logging.NOTSET,
        logging.DEBUG,
        logging.INFO,
        logging.WARN,
        logging.ERROR,
        logging.CRITICAL,
    ],
)
def test_levels(level, build_manager: Callable[..., EpiLog]) -> None:
    """Tests Expected instantiation Behavior of EpiLog class with Logging Level."""
    manager: EpiLog = build_manager(level=level)
    assert manager.level == level, "Expected Logging Level to be reflect input."


def test_level_error(build_manager: Callable[..., EpiLog]) -> None:
    """Tests ValueError is raised Expected when instantiating invalid Logging Level."""
    with pytest.raises(ValueError):
        build_manager(level=-1)


@pytest.mark.parametrize(
    "handler",
    [
        logging.FileHandler("test.log"),
        logging.StreamHandler(),
    ],
)
def test_handlers(handler, build_manager: Callable[..., EpiLog]) -> None:
    """Tests Expected instantiation Behavior of EpiLog class with stream."""
    manager: EpiLog = build_manager(stream=handler)

    # Cleanup by removing file created by file handler
    if isinstance(handler, logging.FileHandler):
        os.remove(handler.baseFilename)

    assert manager.stream == handler, "Expected Stream to reflect input."


def test_handlers_error(build_manager: Callable[..., EpiLog]) -> None:
    """Tests TypeError is raised Expected when instantiating invalid stream."""
    with pytest.raises(TypeError):
        build_manager(stream="handler")


def test_get_log_by_name(build_manager: Callable[..., EpiLog]) -> None:
    """Tests the special dunder method __get_item__ works as expected."""
    manager: EpiLog = build_manager()
    name = "get_item"
    log: logging.Logger = manager.get_logger(name)

    assert manager[name] == log, "Expected to retrieve the same logger via get item."


def _confirm_removal(cls: EpiLog, name: Union[str, logging.Logger]) -> None:
    if isinstance(name, logging.Logger):
        _name = name.name
    else:
        _name = name

    assert _name in cls.loggers, "Expected logger to be registered"
    assert (
        _name in logging.Logger.manager.loggerDict
    ), "Expected logger to be in logging module registry"

    cls.remove(name)

    assert _name not in cls.loggers, "Expected logger to be removed locally"
    assert (
        _name not in logging.Logger.manager.loggerDict
    ), "Expected logger to be removed from logging module registry"

    # Confirm current stream is not closed
    if hasattr(cls.stream, "_closed"):
        # only available >=python3.10
        assert not cls.stream._closed, "Expected current stream to remain open."
    if hasattr(cls.stream, "stream"):
        assert not cls.stream.stream.closed, "Expected current stream to remain open."


def test_log_removal_by_name(build_manager: Callable[..., EpiLog]) -> None:
    """Test we can remove a logger by providing the name of the logger."""
    manager: EpiLog = build_manager()
    name = "removal_by_name"
    manager.get_logger(name)
    _confirm_removal(manager, name)


def test_log_removal_by_logger(build_manager: Callable[..., EpiLog]) -> None:
    """Test we can remove a logger by providing the logger itself."""
    manager: EpiLog = build_manager()
    name = "removal_by_logger"
    log: logging.Logger = manager.get_logger(name)
    _confirm_removal(manager, log)



"""Test Expected Behavior of the EpiLog Benchmark Module."""

import logging
from io import StringIO
from typing import Generator, Tuple

import pytest

from EpiLog import EpiLog
from EpiLog.benchmark import NS_UNITS, BenchMark


@pytest.fixture
def construct(build_manager) -> Generator[Tuple[StringIO, EpiLog], None, None]:
    """Construct EpiLog Log Manager with standard Fixings."""
    stream = StringIO()
    handler = logging.StreamHandler(stream)
    form = logging.Formatter("%(levelname)s | %(message)s")
    manager = build_manager(stream=handler, formatter=form)

    yield stream, manager

    stream.close()
    assert stream.closed, "Expected Stream to be Closed."


def test_empty_benchmark(construct: Tuple[StringIO, EpiLog]) -> None:
    """Tests that an Empty Benchmark context manager class correctly logs message."""
    stream, manager = construct
    manager.level = logging.INFO

    log = manager.get_logger("test")
    msg = "I'm positively bedeviled with meetings et cetera"
    expected = f"INFO | {msg}"

    with BenchMark(log, msg) as b:
        assert b.enabled, "Expected Logging to be Enabled on Benchmark Class."

    stream.seek(0)
    output = stream.read()

    assert msg in output, "Message not Found in output stream after logging."
    assert expected in output, "Message Format not found in Output Stream."
    assert "Traceback" not in output, "Error raised during use of context manager."


def test_enabled(construct: Tuple[StringIO, EpiLog]) -> None:
    """Tests that a Benchmark Logging Level below a Log Level does not log."""
    stream, manager = construct
    manager.level = logging.CRITICAL

    log = manager.get_logger("disabled")
    msg = "That's exactly the kind of paranoia that makes me weary of spending time "
    msg += "with you."

    with BenchMark(log, msg, logging.DEBUG) as b:
        assert not b.enabled, "Expected Logging to be Disabled on Benchmark Class."

    stream.seek(0)
    output = stream.read()
    assert msg not in output, "Message Found in output stream after disabled Benchmark."


def test_error(construct: Tuple[StringIO, EpiLog]) -> None:
    """Test that an error message is emitted when one occurs during context."""
    stream, manager = construct
    manager.level = logging.DEBUG
    log = manager.get_logger("errors")

    with pytest.raises(AssertionError):
        with BenchMark(log, "msg", logging.DEBUG):
            assert False, "We are Intentionally raising an error"  # noqa: B011

    stream.seek(0)
    output = stream.read()
    assert "ERROR" in output, "Expected Error Level Log Emitted"
    assert "Traceback" in output, "Expected Log message to include traceback info."


@pytest.mark.parametrize(
    ["value", "expected"],
    [
        (50, {"ns": 50}),
        (1000, {"us": 1}),
        (1_000_000, {"ms": 1}),
        (5_050_000, {"ms": 5, "us": 50}),
        (5_050_000, {"ms": 5, "us": 50}),
        (1_005_050_000, {"s": 1, "ms": 5, "us": 50}),
    ],
)
def test_breakdown(value: int, expected: dict) -> None:
    """Test Breakdown of ns into appropriate time bins."""
    result = NS_UNITS.breakdown_units(value)
    container = dict((i.unit, 0) for i in NS_UNITS)
    container.update(expected)

    assert result == container, "Expected Equal Output."



# MIT License

# Copyright (c) 2023 Spill-Tea

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Simple context manager to log procedure operation duration, for benchmark."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from time import perf_counter_ns
from typing import Dict, Iterator, Tuple, Union


@dataclass
class Unit:
    """Unit comparing string id to a modifier of the next unit.

    Args:
        unit (str): name of unit
        base (int): value of unit
        modifier (int | None): multiplier describing 1 of next unit

    """

    unit: str
    base: int = 1
    modifier: Union[int, None] = None


class Units:
    """Container of Units.

    Args:
        units (Unit): unit defintions

    """

    units: Tuple[Unit, ...]

    def __init__(self, *units: Unit):
        self.units = units
        self._update()

    def _update(self):
        for n, current in enumerate(self.units[1:]):
            previous = self.units[n]
            current.base = previous.base * previous.modifier

    def __iter__(self) -> Iterator[Unit]:
        yield from self.units

    def convert_units(self, value: int) -> Tuple[float, str]:
        """Convert base unit into most relevant unit."""
        new_time: float = float(value)
        text: str = ""

        for u in self:
            new_time = value / u.base
            text = u.unit
            if u.modifier is None or new_time < u.modifier:
                break

        return new_time, text

    def breakdown_units(self, value: int) -> Dict[str, int]:
        """Split base value into component unit bins."""
        data: Dict[str, int] = {}
        for unit in reversed(self.units):
            data[unit.unit], value = divmod(value, unit.base)

        return data


NS_UNITS = Units(
    Unit(unit="ns", modifier=1000),
    Unit(unit="us", modifier=1000),
    Unit(unit="ms", modifier=1000),
    Unit(unit="s", modifier=60),
    Unit(unit="min", modifier=60),
    Unit(unit="hr", modifier=24),
    Unit(unit="days", modifier=7),
    Unit(unit="weeks"),
)


class BenchMark:
    """Context Manager to Benchmark any process through a log.

    Args:
        log (logging.Logger):
        description (str): Message used to describe actions performed during benchmark.
        level (int): Logging Level

    Attributes:
        enabled (bool): If Benchmark level is compatible with log level to emit message.
        t0 (int): Entry time to benchmark suite.

    Examples:
        ```python
        import logging
        from EpiLog import Benchmark
        log: logging.Logger = logging.getLogger("name")
        message: str = "this is a message"
        with Benchmark(log, message, logging.INFO):
            perform_task(...)
        ```

    """

    __slots__ = ("description", "enabled", "level", "log", "t0")

    level: int
    enabled: bool
    log: logging.Logger
    description: str
    t0: int

    def __init__(
        self,
        log: logging.Logger,
        description: str,
        level: int = logging.INFO,
    ) -> None:
        self.level = level
        self.enabled = log.isEnabledFor(self.level)
        self.log = log
        self.description = description
        self.t0 = 0

    def __enter__(self) -> "BenchMark":
        if self.enabled:
            self.t0 = perf_counter_ns()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        end: int = perf_counter_ns()

        if exc_type is not None:
            self.log.error("Traceback:", exc_info=(exc_type, exc_val, exc_tb))
            return

        if not self.enabled:
            return

        elapsed, unit = NS_UNITS.convert_units(end - self.t0)
        self.log.log(self.level, "%s: (%.4f %s)", self.description, elapsed, unit)
